require "flock/sandbox/version"
require 'open-uri'
require 'erb'
require 'docker'

module Flock
  module Sandbox
    def self.define(&blk)
      @@current_runner.define(&blk)
    end

    def self.set_runner(runner)
      @@current_runner = runner
    end
    class Runner
      attr_accessor :config_path, :target, :definition
      def initialize(args)
        self.config_path, self.target = args
      end

      def run
        instance_eval(File.open(config_path).read, config_path)
        case target
        when "docker"
          Docker.new(definition).run
        when "vagrant"
          Vagrant.new(definition).run
        when "docker-stop"
          DockerStop.new(definition).run
        end
      end

      def define(&blk)
        self.definition = Definition.new
        context = Context.new(self, definition)
        context.instance_exec(definition, &blk)
      end
    end
    class Definition
      attr_accessor :config, :containers, :scripts
      def initialize
        @scripts = []
        @containers = []
      end
      def container(name, params = {})
        @containers << params.merge(name: name)
      end
      def script(name, params = {})
        @scripts << params.merge(name: name)
      end
    end
    class Context
      attr_accessor :runner, :definition
      def initialize(runner, definition)
        self.runner = runner
        self.definition = definition
      end
      def host(container)
        case runner.target
        when "docker"
          container
        when "vagrant"
          definition.config[:private_network]
        end
      end
      def consul_key(key)
        open(definition.config[:consul] + "/v1/kv/" + key + "?raw").read
      end
    end

    class Docker
      def initialize(definition)
        @definition = definition
        ::Docker.options = {}
      end

      def run
        containers = []
        begin
          unless ENV['ONLY_SCRIPTS']
            @definition.containers.each do |c|
              puts "Start container #{c[:name]} (#{c[:image]})"
              ::Docker::Image.create('fromImage' => c[:image])
              container = ::Docker::Container.create(
                'Image' => c[:image],
                'name' => c[:name],
                'Env' => c[:envs],
                'HostConfig' => { 'Links' => c[:links] })
              containers << container
              container.start
            end
          end
          if ENV['MANUAL_SCRIPTS']
            @definition.scripts.each do |c|
              puts "docker run #{(c[:links] || [])
                  .map { |v| "--link #{v}" } * ' ' } #{(c[:envs] || [])
                  .map { |e| "-e #{e}" } * ' ' } #{c[:image]} #{c[:cmd] * ' '}"
            end
          else
            @definition.scripts.each do |c|
              puts "Run script #{c[:name]} (#{c[:image]} #{c[:cmd].inspect})"
              ::Docker::Image.create('fromImage' => c[:image])
              container = ::Docker::Container.create(
                'Cmd' => c[:cmd],
                'Image' => c[:image],
                'Env' => c[:envs],
                'HostConfig' => { 'Links' => c[:links] })
              containers << container
              container.start
              res = container.wait(1200)
              logs = container.logs(stdout: true, stderr: true)
              puts logs
              if res['StatusCode'] != 0
                puts "Error in script #{c[:name]}: error code #{res['StatusCode']}"
                exit res['StatusCode']
              end
            end
          end
        ensure
          unless ENV['MANUAL_SCRIPTS']
            containers.each do |c|
              c.kill
              c.delete
            end
          end
        end
      end
    end

    class DockerStop

      def initialize(definition)
        @definition = definition
      end

      def run
        @definition.containers.each do |c|
          cont = begin
                   ::Docker::Container.get(c[:name])
                 rescue ::Docker::Error::NotFoundError
                 end
          next unless c
          puts "Stop #{c[:name]}"
          cont.kill
          cont.delete
        end
      end
    end

    class Vagrant
      def initialize(definition)
        @definition = definition
      end

      def run
        template = ERB.new(File.read(File.expand_path('../../../resources/vagrant.erb', __FILE__)))
        puts template.result(@definition.instance_eval { binding })
      end
    end
  end
end
